"""
styles - constants and functions for CSS styles
"""

from .shared import named_styles, style_def

hover_style_name = "psc-hover-data"

default_styles = {
    f".psc-hover-group .{hover_style_name}": {
        "display": "none",
    },
    f".psc-hover-group:hover .{hover_style_name}": {
        "display": "inline",
    },
}


def render_style_dict(style_dict: style_def) -> str:
    indent = " " * 4
    nl = "\n"
    return f"""
{{
{indent}{f"{nl}{indent}".join(f"{param}: {value};" for param, value in style_dict.items())}
}}
    """.strip()


def optionally_merge_styles_to_default(
    styles: named_styles | None = None,
    include_default: bool = True,
) -> named_styles:
    if styles is None:
        return {**default_styles} if include_default else {}
    return {**styles, **default_styles} if include_default else {**styles}


def render_all_styles(
    styles: named_styles | None = None,
    include_default: bool = True,
) -> str:
    rendered_styles = (
        default_styles.copy()
        if styles is None
        else optionally_merge_styles_to_default(styles, include_default)
    )
    return "\n".join(
        [f"{name} {render_style_dict(rendered_styles[name])}\n" for name in rendered_styles]  # type: ignore[operator]
    )[:-1]



"""Top-level package for py-svg-chart"""

__author__ = "Alex Rowley"
__email__ = ""
__version__ = "0.5.6"

from .charts import (
    BarChart,
    DonutChart,
    LineChart,
    NormalisedBarChart,
    ScatterChart,
    SimpleLineChart,
)
from .shapes import (
    Circle,
    Line,
    Text,
)
from .styles import (
    hover_style_name,
    render_all_styles,
)



from __future__ import annotations
from abc import abstractmethod
from typing import Any, Callable

from .helpers import collapse_element_list
from .scales import make_linear_scale
from .shapes import Shape, Text, Line
from .shared import number, style_def


class Axis(Shape):
    """
    axis of a graph
    """

    default_axis_styles: style_def = {"stroke": "#2e2e2c"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        min_unique_values: int = 2,
        scale_maker=make_linear_scale,
        secondary: bool = False,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        _ignore = secondary, axis_styles, tick_length
        super().__init__(x_position, y_position)
        self.data_points = data_points
        self.length = axis_length
        self.scale = scale_maker(
            values=data_points,
            max_ticks=max_ticks,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            min_unique_values=min_unique_values,
        )
        self.label_format = label_format
        self.axis_line: Line | None = None
        self.tick_lines: list[Line] = []
        self.tick_texts: list[Text] = []
        self.grid_lines: list[Line] = []
        self.title: Text | None = None

    def get_element_list(self):
        return collapse_element_list(
            [self.axis_line, self.title] if self.title else [self.axis_line],
            self.tick_lines,
            self.tick_texts,
            self.grid_lines,
        )

    @abstractmethod
    def get_positions(self, values, include_axis=True) -> list[int | float | None]: ...

    def get_ticks_with_positions(self) -> list[tuple[Any, int | float | None]]:
        return list(zip(self.scale.ticks, self.get_positions(self.scale.ticks)))


class XAxis(Axis):
    """
    x-axis of a graph
    """

    default_tick_text_styles = {"text-anchor": "middle", "dominant-baseline": "hanging"}
    default_title_styles = {"text-anchor": "middle", "dominant-baseline": "middle"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        scale_maker=make_linear_scale,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        super().__init__(
            x_position=x_position,
            y_position=y_position,
            data_points=data_points,
            axis_length=axis_length,
            label_format=label_format,
            max_ticks=max_ticks,
            axis_styles=axis_styles,
            tick_length=tick_length,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            min_unique_values=2,  # at least two unique values needed on the x-axis to create a meaningful graph
            scale_maker=scale_maker,
        )
        styles = axis_styles or self.default_axis_styles.copy()
        self.axis_line = Line(
            x=self.position.x,
            y=self.position.y,
            width=axis_length,
            height=0,
            styles=styles,
        )

        for tick, pos in self.get_ticks_with_positions():
            if pos is None:  # shifted out of the visible range
                continue
            self.tick_lines.append(
                Line(x=pos, width=0, y=self.position.y, height=tick_length, styles=styles),
            )
            self.tick_texts.append(
                Text(
                    x=pos,
                    y=self.position.y + 2 * tick_length,
                    content=label_format(tick),
                    styles=self.default_tick_text_styles.copy(),
                ),
            )

        if title:
            title_x = self.position.x + self.length / 2  # Center on the axis
            title_y = self.position.y + title_offset
            self.title = Text(
                x=title_x,
                y=title_y,
                content=title,
                styles=title_styles or self.default_title_styles.copy()
            )

    def get_positions(self, values, include_axis=True) -> list[int | float | None]:
        proportions_of_range = [self.scale.value_to_fraction(value) for value in values]
        in_range = (
            (lambda prop: 0.0 <= prop <= 1.0) if include_axis else (lambda prop: 0.0 < prop <= 1.0)
        )
        return [
            self.position.x + prop * self.length if in_range(prop) else None
            for prop in proportions_of_range
        ]


class YAxis(Axis):
    """
    y-axis of a graph
    """

    default_tick_text_styles = {"text-anchor": "end", "dominant-baseline": "middle"}
    default_sec_tick_text_styles = {"text-anchor": "start", "dominant-baseline": "middle"}
    default_title_styles = {"text-anchor": "middle", "dominant-baseline": "middle"}

    def __init__(
        self,
        x_position: number,
        y_position: number,
        data_points,
        axis_length: number,
        label_format: Callable,
        max_ticks: int = 10,
        axis_styles: style_def | None = None,
        tick_length: int = 5,
        min_value=None,
        max_value=None,
        include_zero: bool = False,
        shift: bool = False,
        scale_maker=make_linear_scale,
        secondary: bool = False,
        title: str | None = None,
        title_styles: style_def | None = None,
        title_offset: int = 40
    ):
        super().__init__(
            x_position=x_position,
            y_position=y_position,
            data_points=data_points,
            axis_length=axis_length,
            label_format=label_format,
            max_ticks=max_ticks,
            axis_styles=axis_styles,
            tick_length=tick_length,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
            shift=shift,
            secondary=secondary,
            min_unique_values=1,  # one unique value is sufficient for the y-axis
            scale_maker=scale_maker,
        )
        styles = axis_styles or self.default_axis_styles.copy()
        self.axis_line = Line(
            x=self.position.x,
            y=self.position.y,
            width=0,
            height=axis_length,
            styles=styles,
        )

        if secondary:
            tick_pos_offset = 0
            tick_text_offset = 2 * tick_length
            tick_text_styles = self.default_sec_tick_text_styles.copy()
        else:
            tick_pos_offset = -tick_length
            tick_text_offset = -2 * tick_length
            tick_text_styles = self.default_tick_text_styles.copy()

        for tick, pos in self.get_ticks_with_positions():
            if pos is None:  # shifted out of the visible range
                continue
            self.tick_lines.append(
                Line(
                    x=self.position.x + tick_pos_offset,
                    y=pos,
                    width=tick_length,
                    height=0,
                    styles=styles,
                ),
            )
            self.tick_texts.append(
                Text(
                    x=self.position.x + tick_text_offset,
                    y=pos,
                    content=label_format(tick),
                    styles=tick_text_styles,
                ),
            )

        if title:
            title_x = self.position.x + tick_text_offset + (title_offset if secondary else -title_offset)
            title_y = self.position.y + self.length / 2
            styles = title_styles or self.default_title_styles.copy()
            styles['transform'] = f'rotate(-90 {title_x} {title_y})'
            self.title = Text(
                x=title_x,
                y=title_y,
                content=title,
                styles=styles
            )

    def get_positions(self, values, include_axis=True) -> list[int | float | None]:
        proportions_of_range = [1 - self.scale.value_to_fraction(value) for value in values]
        in_range = (
            (lambda prop: 0.0 <= prop <= 1.0) if include_axis else (lambda prop: 0.0 <= prop < 1.0)
        )
        return [
            self.position.y + prop * self.length if in_range(prop) else None
            for prop in proportions_of_range
        ]



"""
shared - definitions shared throughout the code
"""

from datetime import date, datetime


dates_sequence = list[date] | tuple[date, ...]
datetimes_sequence = list[datetime] | tuple[datetime, ...]

number = float | int
numbers_sequence = list[number] | tuple[number, ...]

style_def = dict[str, str]
named_styles = dict[str, style_def]



from typing import Callable
import math

from .helpers import collapse_element_list
from .shapes import Circle, Element, Point, Shape
from .shared import number, numbers_sequence, style_def


class Series(Shape):
    """
    base class for series
    """

    def __init__(
            self,
            x_position: number,
            y_position: number,
            styles: style_def | None = None,
            classes: list[str] | None = None,
    ):
        super().__init__(x=x_position, y=y_position, styles=styles, classes=classes)
        self.custom_elements: list[Element] = []

    @property
    def pv_generator(self):
        return []

    def add_custom_elements(self, custom_elements: list[Element]):
        self.custom_elements.extend(custom_elements)

    def get_element_list(self) -> list:
        return []


class DonutSegment(Series):
    """
    donut chart segment
    """

    # fmt: off
    path_template = (
        '<path d="M {outer_begin_x},{outer_begin_y} '
        'A {radius_outer} {radius_outer} 0 {large_arc_flag} 1 {outer_end_x} {outer_end_y} '
        'L {inner_begin_x},{inner_begin_y} '
        'A {radius_inner} {radius_inner} 0 {large_arc_flag} 0 {inner_end_x} {inner_end_y} '
        'Z" {attributes}></path>'
    )

    # fmt: on

    def __init__(
            self,
            colour: str | number,
            start_theta: number,
            end_theta: number,
            radius_inner: number,
            radius_outer: number,
            centre_x: number,
            centre_y: number,
            styles: style_def | None = None,
            classes: list[str] | None = None,
    ):
        super().__init__(x_position=centre_x, y_position=centre_y, styles=styles, classes=classes)
        self.start_theta = start_theta
        self.end_theta = end_theta
        self.centre_x = centre_x
        self.centre_y = centre_y
        self.radius_inner = radius_inner
        self.radius_outer = radius_outer
        self.styles = {"fill": colour}

    @property
    def start_theta_rad(self):
        return math.radians(self.start_theta)

    @property
    def end_theta_rad(self):
        return math.radians(self.end_theta)

    @property
    def inner_begin_x(self):
        return self.position.x + self.radius_inner * math.cos(self.end_theta_rad)

    @property
    def inner_end_x(self):
        return self.position.x + self.radius_inner * math.cos(self.start_theta_rad)

    @property
    def inner_begin_y(self):
        return self.position.y + self.radius_inner * math.sin(self.end_theta_rad)

    @property
    def inner_end_y(self):
        return self.position.y + self.radius_inner * math.sin(self.start_theta_rad)

    @property
    def outer_begin_x(self):
        return self.position.x + self.radius_outer * math.cos(self.start_theta_rad)

    @property
    def outer_end_x(self):
        return self.position.x + self.radius_outer * math.cos(self.end_theta_rad)

    @property
    def outer_begin_y(self):
        return self.position.y + self.radius_outer * math.sin(self.start_theta_rad)

    @property
    def outer_end_y(self):
        return self.position.y + self.radius_outer * math.sin(self.end_theta_rad)

    @property
    def large_arc_flag(self):
        return 1 if (self.end_theta - self.start_theta) > 180 else 0

    def get_element_list(self) -> list:
        return [
            self.path_template.format(
                outer_begin_x=self.outer_begin_x,
                outer_begin_y=self.outer_begin_y,
                radius_inner=self.radius_inner,
                radius_outer=self.radius_outer,
                large_arc_flag=self.large_arc_flag,
                outer_end_x=self.outer_end_x,
                outer_end_y=self.outer_end_y,
                inner_end_x=self.inner_end_x,
                inner_end_y=self.inner_end_y,
                inner_begin_x=self.inner_begin_x,
                inner_begin_y=self.inner_begin_y,
                attributes=self.attributes,
            )
        ] + collapse_element_list(self.custom_elements)


class LineSeries(Series):
    """
    line series given as a number of (x, y)-points
    """

    __default_styles__ = {"stroke-width": "2"}
    path_begin_template = '<path d="{path}" fill="none" {attributes}/>'

    def __init__(
            self,
            points: list[Point],
            x_values: numbers_sequence,
            y_values: numbers_sequence,
            styles: style_def | None = None,
            classes: list[str] | None = None,
    ):
        super().__init__(
            x_position=points[0].x,
            y_position=points[0].y,
            styles=styles,
            classes=classes,
        )
        self.points = points
        self.x_values = x_values
        self.y_values = y_values

    @property
    def pv_generator(self):
        return zip(self.points, self.x_values, self.y_values)

    @property
    def path_length(self) -> number:
        return (
            sum(
                math.sqrt((p2.x - p1.x) ** 2 + (p2.y - p1.y) ** 2)
                for p1, p2 in zip(self.points, self.points[1:])
            )
            if len(self.points) > 1
            else 0
        )

    def get_element_list(self) -> list:
        path = " ".join(
            [f"L {p.x} {p.y}" if i else f"M {p.x} {p.y}" for i, p in enumerate(self.points)]
        )
        return [
            self.path_begin_template.format(path=path, attributes=self.attributes)
        ] + collapse_element_list(self.custom_elements)


class BarSeries(Series):
    """
    series for bar charts
    """

    bar_template = '<rect x="{x}" y="{y}" width="{w}" height="{h}" {attributes}/>'
    __default_styles__ = {"stroke": "none"}

    def __init__(
            self,
            points: list[Point],
            x_values: numbers_sequence,
            y_values: numbers_sequence,
            bar_width: number,
            bar_heights: numbers_sequence,
            styles: style_def | None = None,
            classes: list[str] | None = None,
    ):
        super().__init__(
            x_position=points[0].x,
            y_position=points[0].y,
            styles=styles,
            classes=classes,
        )
        self.points = points
        self.x_values = x_values
        self.y_values = y_values
        self.bar_width = bar_width
        self.bar_heights = bar_heights

    @property
    def pv_generator(self):
        return zip(self.points, self.x_values, self.y_values)

    def get_element_list(self) -> list:
        bars = [
            self.bar_template.format(
                x=p.x - self.bar_width / 2,
                y=p.y,
                w=self.bar_width,
                h=h,
                attributes=self.attributes,
            )
            for p, h in zip(self.points, self.bar_heights)
        ]
        return bars + collapse_element_list(self.custom_elements)


def default_scatter_shape_template(
        x: number,
        y: number,
        styles: style_def,
) -> Shape:
    return Circle(x, y, radius=3, styles=styles)


class ScatterSeries(Series):
    """
    scatter series given as a number of (x, y)-points
    """

    __default_styles__: style_def = {}
    __default_shape_template__ = staticmethod(default_scatter_shape_template)

    def __init__(
            self,
            points: list[Point],
            x_values: numbers_sequence,
            y_values: numbers_sequence,
            shape_template: Callable[[number, number, style_def], Shape] | None = None,
            styles: style_def | None = None,
            classes: list[str] | None = None,
    ):
        super().__init__(
            x_position=points[0].x,
            y_position=points[0].y,
            styles=styles,
            classes=classes,
        )
        self.points = points
        self.x_values = x_values
        self.y_values = y_values
        self.shape_template = (
            self.__default_shape_template__ if shape_template is None else shape_template
        )

    @property
    def pv_generator(self):
        return zip(self.points, self.x_values, self.y_values)

    def get_element_list(self) -> list:
        return collapse_element_list(
            [self.shape_template(p.x, p.y, self.styles) for p in self.points]
        ) + collapse_element_list(self.custom_elements)



import math
import datetime as dt


from .shared import (
    dates_sequence,
    datetimes_sequence,
    number,
    numbers_sequence,
)


def noop(*args, **kwargs) -> None:
    pass


def default_format(value) -> str:
    """
    format a number
    """
    return "{0:,}".format(value) if not isinstance(value, str) else value


def safe_get_element_list(elements):
    """
    elements may not exist and may lack the get_element_list() function
    """
    if elements is not None and hasattr(elements, "get_element_list"):
        yield from elements.get_element_list()


def collapse_element_list(*list_of_list_of_elements) -> list[str]:
    """
    flatten any number of lists of elements to a list of elements
    """
    return [
        element
        for list_of_elements in list_of_list_of_elements
        for elements in (
            list_of_elements if isinstance(list_of_elements, (list, tuple, set)) else []
        )
        for element in safe_get_element_list(elements)
    ]


def get_numeric_ticks(
    values: numbers_sequence,
    max_ticks: int,
    min_value: number | None = None,
    max_value: number | None = None,
    include_zero: bool = False,
) -> numbers_sequence:
    """
    compute ticks for a series of numbers
    :param values: actual values
    :param max_ticks: maximum number of ticks
    :param min_value: optional minimum value to include in ticks
    :param max_value: optional maximum value to include in ticks
    :param include_zero: whether to include zero in ticks
    """
    value_min, value_max = min(values), max(values)
    if min_value is not None:
        value_min = min(value_min, min_value)
    if max_value is not None:
        value_max = max(value_max, max_value)
    if include_zero:
        if value_min > 0:
            value_min = 0
        if value_max < 0:
            value_max = 0

    if value_max == value_min:
        raise ValueError("All values are the same — cannot compute min/max.")

    raw_pad = (value_max - value_min) / max_ticks
    magnitude = 10 ** math.floor(math.log10(raw_pad))
    residual = raw_pad / magnitude
    if residual <= 1:
        nice = 1
    elif residual <= 2:
        nice = 2
    elif residual <= 5:
        nice = 5
    else:
        nice = 10

    pad = nice * magnitude
    start = math.floor(value_min / pad)
    end = math.ceil(value_max / pad)
    return [round(y * pad, 10) for y in range(int(start), int(end + 1))]


def get_logarithmic_ticks(
    values: numbers_sequence,
    max_ticks: int,
    min_value: number | None = None,
    max_value: number | None = None,
    include_zero: bool = False,
) -> numbers_sequence:
    """
    compute logarithmic ticks for a series of numbers
    :param values: actual values
    :param max_ticks: maximum number of ticks
    :param min_value: optional minimum value to include in ticks
    :param max_value: optional maximum value to include in ticks
    :param include_zero: whether to include zero in ticks
    """
    value_min, value_max = min(values), max(values)
    if min_value is not None:
        value_min = min(value_min, min_value)
    if max_value is not None:
        value_max = max(value_max, max_value)
    if include_zero:
        if value_min > 0:
            value_min = 0
        if value_max < 0:
            value_max = 0

    if value_max == value_min:
        raise ValueError("All values are the same — cannot compute min/max.")

    start = math.floor(math.log10(value_min))
    end = math.ceil(math.log10(value_max)) + 1  # optimization, as we only need end+1
    step = 1
    while (end - start) // step > max_ticks:
        step *= 2
    if 10 ** ((end - start) // step) < value_max:
        end += 1
        step += 1
    return [10**y for y in range(int(start), int(end), int(step))]


def get_date_or_time_ticks(
    dates: dates_sequence | datetimes_sequence,
    max_ticks: int = 10,
    min_value: dt.date | dt.datetime | None = None,
    max_value: dt.date | dt.datetime | None = None,
) -> dates_sequence | datetimes_sequence:
    """
    compute ticks for a series of dates/datetimes
    :param dates: actual dates/datetimes
    :param max_ticks: maximum number of ticks
    :param min_value: optional minimum value to include in ticks
    :param max_value: optional maximum value to include in ticks
    """
    date_min, date_max = min(dates), max(dates)
    if date_min >= date_max:
        raise ValueError("Dates must have a positive range.")

    if min_value and min_value < date_min:
        date_min = min_value

    if max_value and max_value > date_max:
        date_max = max_value

    total_seconds = (date_max - date_min).total_seconds()

    if total_seconds <= 300:  # 5 minutes
        interval = dt.timedelta(seconds=max(1, int(total_seconds // max_ticks)))
    elif total_seconds <= 3600:  # an hour
        interval = dt.timedelta(minutes=max(1, int(total_seconds // (max_ticks * 60))))
    elif total_seconds <= 86400:  # a day
        interval = dt.timedelta(hours=max(1, int(total_seconds // (max_ticks * 3600))))
    elif total_seconds <= 30 * 86400:  # about a month
        interval = dt.timedelta(days=max(1, int(total_seconds // (max_ticks * 86400))))
    else:
        total_days = total_seconds / 86400
        approx_months = total_days / 30.0
        raw_interval = approx_months / max_ticks

        if raw_interval <= 1:
            interval_months = 1
        elif raw_interval <= 2:
            interval_months = 2
        elif raw_interval <= 3:
            interval_months = 3
        elif raw_interval <= 6:
            interval_months = 6
        else:
            interval_months = 12

        # enclosing period of whole months
        start = date_min.replace(day=1)
        end = (date_max.replace(day=1) + dt.timedelta(days=32)).replace(day=1)
        # time to 00:00:00 for datetimes
        if isinstance(start, dt.datetime):
            start = dt.datetime.combine(start.date(), dt.time(0, 0, 0))
        if isinstance(end, dt.datetime):
            end = dt.datetime.combine(end.date(), dt.time(0, 0, 0))

        current_tick = start
        ticks = [current_tick]
        while ticks[-1] < end:
            month = current_tick.month + interval_months
            year = current_tick.year + (month - 1) // 12
            month = (month - 1) % 12 + 1
            current_tick = current_tick.replace(year=year, month=month)
            ticks.append(current_tick)
        return ticks

    current_tick = (
        date_min.replace(second=0, microsecond=0) if isinstance(date_min, dt.datetime) else date_min
    )
    ticks = [current_tick]
    while ticks[-1] < date_max:
        current_tick += interval
        ticks.append(current_tick)

    return ticks



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime, date, timedelta
from typing import Any

import math

from .helpers import (
    get_date_or_time_ticks,
    get_logarithmic_ticks,
    get_numeric_ticks,
)
from .shared import (
    dates_sequence,
    datetimes_sequence,
    number,
    numbers_sequence,
)


class Scale(ABC):
    """
    base class for scales
    """

    def __init__(self, ticks: Any):
        self.ticks = ticks

    @abstractmethod
    def get_lowest(self) -> Any:
        """
        get lowest value iff meaningful, None otherwise
        """
        ...

    @abstractmethod
    def value_to_fraction(self, value) -> float:
        """
        proportion of the scale where the value is positioned: [lo; hi] -> [0.0; 1.0]
        NOTE outside [0.0; 1.0] means the value is outside the scale.
        """
        ...


class MappedLinearScale(Scale):
    """
    mapped linear scale - evenly spaced mapped value scale
    """

    lo: date | datetime | number
    hi: date | datetime | number
    size: timedelta | number

    def __init__(
        self,
        ticks: dates_sequence | datetimes_sequence | numbers_sequence,
        shift: bool | date | datetime | number = False,
    ):
        if all(isinstance(tick, date) for tick in ticks):
            pass
        elif all(isinstance(tick, datetime) for tick in ticks):
            pass
        elif all(isinstance(tick, number) for tick in ticks):
            pass
        else:
            raise TypeError("LinearScale only supports date, datetime, float/int values")
        super().__init__(ticks)
        self.lo = self.map_value(min(ticks))
        self.hi = self.map_value(max(ticks))
        self.size = self.hi - self.lo  # type: ignore[operator]
        if shift is False:
            self.shift = None
        elif isinstance(self.lo, date) and isinstance(shift, date):
            self.shift = (self.map_value(shift) - self.lo) / self.size  # type: ignore[operator]
        elif isinstance(self.lo, datetime) and isinstance(shift, datetime):
            self.shift = (self.map_value(shift) - self.lo) / self.size  # type: ignore[operator]
        elif isinstance(self.lo, number) and isinstance(shift, number):
            self.shift = (self.map_value(shift) - self.lo) / self.size  # type: ignore[operator]
        else:
            self.shift = None

    def __str__(self):
        return f"[{self.lo}...{self.hi}] {self.shift if self.shift else '(no shift)'}"

    def __repr__(self):
        return f"<{self.__class__.__name__} lo={self.lo} hi={self.hi} size={self.size} shift={self.shift}>"

    def get_lowest(self) -> date | datetime | number:
        return self.lo

    def value_to_fraction(self, value: date | datetime | number) -> float:
        fraction = (self.map_value(value) - self.lo) / self.size  # type: ignore[operator]
        return fraction - self.shift if self.shift else fraction  # type: ignore[operator, return-value]

    @staticmethod
    @abstractmethod
    def map_value(value: date | datetime | number) -> date | datetime | number: ...


class LinearScale(MappedLinearScale):
    """
    linear scale - values unchanged in mapping
    """

    @staticmethod
    def map_value(value: date | datetime | number) -> date | datetime | number:
        """
        values unchanged in mapping
        """
        return value


class LogarithmicScale(MappedLinearScale):
    """
    logarithmic scale - log10 to map values
    """

    def __init__(
        self,
        ticks: numbers_sequence,
        shift: bool | number = False,
    ):
        if not all(isinstance(tick, number) for tick in ticks):
            raise TypeError("LogarithmicScale only supports float/int values")
        super().__init__(ticks, shift)

    @staticmethod
    def map_value(value: number) -> number:  # type: ignore[override]
        """
        values unchanged in mapping
        """
        return math.log10(value)


class MappingScale(Scale):
    """
    scale for non-numeric values
    """

    def __init__(self, ticks: list | tuple):
        super().__init__(ticks)
        value_width = 1.0 / len(ticks)
        self.map = {value: (index + 0.5) * value_width for index, value in enumerate(ticks)}

    def __str__(self):
        return f"""[{", ".join(f"{tick}" for tick in self.ticks)}]"""

    def __repr__(self):
        return f"""<{self.__class__.__name__} [{", ".join(f"{tick}" for tick in self.ticks)}]>"""

    def get_lowest(self) -> None:
        return None

    def value_to_fraction(self, value) -> float:
        return self.map.get(value, -1.0)


def make_categories_scale(
    values: list | tuple,
    max_ticks: int,
    min_value: Any = None,
    max_value: Any = None,
    include_zero: bool = False,
    shift: bool = False,
    min_unique_values: int = 2,
) -> Scale:
    """
    make a categories scale for a series of values

    :param values: actual values
    :param max_ticks: maximum number of ticks on the scale
    :param min_value: optional minimum value to include on the scale
    :param max_value: optional maximum value to include on the scale
    :param include_zero: whether to include zero on the scale
    :param shift: optional shift for the scale
    :param min_unique_values: minimum number of unique values required
    """
    _ignore = max_ticks, min_value, max_value, include_zero, shift, min_unique_values
    return MappingScale(list(values))


def make_logarithmic_scale(
    values: numbers_sequence,
    max_ticks: int,
    min_value: number | None = None,
    max_value: number | None = None,
    include_zero: bool = False,
    shift: bool = False,
    min_unique_values: int = 2,
) -> Scale:
    if (
        values is None
        or not isinstance(values, list | tuple)
        or len(set(values)) < min_unique_values
    ):
        raise ValueError(
            "Values must be non-empty with at least %d unique elements.",
            min_unique_values,
        )
    if all(isinstance(value, int | float) for value in values):
        ticks = get_logarithmic_ticks(
            values,
            max_ticks,
            min_value=min_value,
            max_value=max_value,
            include_zero=include_zero,
        )
        return LogarithmicScale(ticks, shift=min(values) if shift is True else shift)
    # mixed value types or value type for which there's no ticks creator
    return MappingScale(list(values))


def make_linear_scale(
    values: dates_sequence | datetimes_sequence | numbers_sequence,
    max_ticks: int,
    min_value: date | datetime | number | None = None,
    max_value: date | datetime | number | None = None,
    include_zero: bool = False,
    shift: bool = False,
    min_unique_values: int = 2,
) -> Scale:
    """
    make a scale for a series of values

    :param values: actual values
    :param max_ticks: maximum number of ticks on the scale
    :param min_value: optional minimum value to include on the scale
    :param max_value: optional maximum value to include on the scale
    :param include_zero: whether to include zero on the scale
    :param shift: optional shift for the scale
    :param min_unique_values: minimum number of unique values required
    """
    if (
        values is None
        or not isinstance(values, list | tuple)
        or len(set(values)) < min_unique_values
    ):
        raise ValueError(
            "Values must be non-empty with at least %d unique elements.",
            min_unique_values,
        )
    # value types for which there is a ticks creator
    ticks: dates_sequence | datetimes_sequence | numbers_sequence
    if all(isinstance(value, date) for value in values):
        ticks = get_date_or_time_ticks(
            values,  # type: ignore[arg-type]
            max_ticks,
            min_value=min_value,  # type: ignore[arg-type]
            max_value=max_value,  # type: ignore[arg-type]
        )
        return LinearScale(ticks, shift=min(values) if shift is True else shift)
    if all(isinstance(value, datetime) for value in values):
        ticks = get_date_or_time_ticks(
            values,  # type: ignore[arg-type]
            max_ticks,
            min_value=min_value,  # type: ignore[arg-type]
            max_value=max_value,  # type: ignore[arg-type]
        )
        return LinearScale(ticks, shift=min(values) if shift is True else shift)
    if all(isinstance(value, int | float) for value in values):
        ticks = get_numeric_ticks(
            values,  # type: ignore[arg-type]
            max_ticks,
            min_value=min_value,  # type: ignore[arg-type]
            max_value=max_value,  # type: ignore[arg-type]
            include_zero=include_zero,
        )
        return LinearScale(ticks, shift=min(values) if shift is True else shift)
    # mixed value types or value type for which there's no ticks creator
    return MappingScale(list(values))
